"""Main COMBO model."""
from typing import Optional, Dict, Any, List

import numpy
import torch
from overrides import overrides
from torch import Tensor

from combo import data
from combo.config import FromParameters, Registry
from combo.config.from_parameters import register_arguments
from combo.data import Instance
from combo.data.batch import Batch
from combo.data.dataset_loaders.dataset_loader import TensorDict
from combo.modules import TextFieldEmbedder
from combo.modules.lemma import LemmatizerModel
from combo.modules.model import Model
from combo.modules.morpho import MorphologicalFeatures
from combo.modules.parser import DependencyRelationModel
from combo.modules.seq2seq_encoders.seq2seq_encoder import Seq2SeqEncoder
from combo.nn import RegularizerApplicator
from combo.nn import utils
from combo.nn.utils import get_text_field_mask
from combo.predictors import Predictor
from combo.utils import metrics
from combo.utils import ConfigurationError


@Registry.register("semantic_multitask")
class ComboModel(Model, FromParameters):
    """Main COMBO model."""

    @classmethod
    def pass_down_parameter_names(cls) -> List[str]:
        return ['vocabulary', 'serialization_dir', 'model_name']

    @register_arguments
    def __init__(self,
                 vocabulary: data.Vocabulary,
                 loss_weights: Dict[str, float],
                 text_field_embedder: TextFieldEmbedder,
                 seq_encoder: Seq2SeqEncoder,
                 use_sample_weight: bool = True,
                 lemmatizer: LemmatizerModel = None,
                 upos_tagger: MorphologicalFeatures = None,
                 xpos_tagger: MorphologicalFeatures = None,
                 semantic_relation: Predictor = None,
                 morphological_feat: MorphologicalFeatures = None,
                 dependency_relation: DependencyRelationModel = None,
                 enhanced_dependency_relation: DependencyRelationModel = None,
                 regularizer: RegularizerApplicator = None,
                 serialization_dir: Optional[str] = None) -> None:
        super().__init__(vocabulary, regularizer, serialization_dir)

        self.text_field_embedder = text_field_embedder
        self.loss_weights = loss_weights
        self.use_sample_weight = use_sample_weight
        self.seq_encoder = seq_encoder
        self.lemmatizer = lemmatizer
        self.upos_tagger = upos_tagger
        self.xpos_tagger = xpos_tagger
        self.semantic_relation = semantic_relation
        self.morphological_feat = morphological_feat
        self.dependency_relation = dependency_relation
        self.enhanced_dependency_relation = enhanced_dependency_relation
        self._head_sentinel = torch.nn.Parameter(torch.randn([1, 1, self.seq_encoder.get_output_dim()]))
        self.scores = metrics.SemanticMetrics()
        self._partial_losses = None

        self.model_name = getattr(
            getattr(self.text_field_embedder, '_token_embedders', {}).get('token'), 'constructed_args'
        ).get('model_name', '')

    def forward_on_tensors(self,
                           sentence: Dict[str, Dict[str, torch.Tensor]],
                           metadata: List[Dict[str, Any]],
                           upostag: torch.Tensor = None,
                           xpostag: torch.Tensor = None,
                           lemma: Dict[str, Dict[str, torch.Tensor]] = None,
                           feats: torch.Tensor = None,
                           head: torch.Tensor = None,
                           deprel: torch.Tensor = None,
                           semrel: torch.Tensor = None,
                           enhanced_heads: torch.Tensor = None,
                           enhanced_deprels: torch.Tensor = None) -> Dict[str, torch.Tensor]:

        # Prepare masks
        char_mask = sentence["char"]["token_characters"].gt(0)
        word_mask = get_text_field_mask(sentence)

        device = word_mask.device

        # If enabled weight samples loss by log(sentence_length)
        sample_weights = word_mask.sum(-1).float().log() if self.use_sample_weight else None

        encoder_input = self.text_field_embedder(sentence, char_mask=char_mask)
        encoder_emb = self.seq_encoder(encoder_input, word_mask)

        batch_size, _, encoding_dim = encoder_emb.size()

        # Concatenate the head sentinel (ROOT) onto the sentence representation.
        head_sentinel = self._head_sentinel.expand(batch_size, 1, encoding_dim)
        encoder_emb_with_root = torch.cat([head_sentinel, encoder_emb], 1)
        word_mask_with_root = torch.cat([torch.ones((batch_size, 1), device=device), word_mask], 1)

        upos_output = self._optional(self.upos_tagger,
                                     encoder_emb,
                                     mask=word_mask,
                                     labels=upostag,
                                     sample_weights=sample_weights)
        xpos_output = self._optional(self.xpos_tagger,
                                     encoder_emb,
                                     mask=word_mask,
                                     labels=xpostag,
                                     sample_weights=sample_weights)
        semrel_output = self._optional(self.semantic_relation,
                                       encoder_emb,
                                       mask=word_mask,
                                       labels=semrel,
                                       sample_weights=sample_weights)
        morpho_output = self._optional(self.morphological_feat,
                                       encoder_emb,
                                       mask=word_mask,
                                       labels=feats,
                                       sample_weights=sample_weights)
        lemma_output = self._optional(self.lemmatizer,
                                      (encoder_emb, sentence.get("char").get("token_characters")
                                      if sentence.get("char") else None),
                                      mask=word_mask,
                                      labels=lemma.get("char").get("token_characters") if lemma else None,
                                      sample_weights=sample_weights)
        parser_output = self._optional(self.dependency_relation,
                                       encoder_emb_with_root,
                                       returns_tuple=True,
                                       mask=word_mask_with_root,
                                       labels=(deprel, head),
                                       sample_weights=sample_weights)
        enhanced_parser_output = self._optional(self.enhanced_dependency_relation,
                                                encoder_emb_with_root,
                                                returns_tuple=True,
                                                mask=word_mask_with_root,
                                                labels=(enhanced_deprels, head, enhanced_heads),
                                                sample_weights=sample_weights)
        relations_pred, head_pred = parser_output["prediction"]
        enhanced_relations_pred, enhanced_head_pred = enhanced_parser_output["prediction"]
        output = {
            "upostag": upos_output["prediction"],
            "xpostag": xpos_output["prediction"],
            "semrel": semrel_output["prediction"],
            "feats": morpho_output["prediction"],
            "lemma": lemma_output["prediction"],
            "head": head_pred,
            "deprel": relations_pred,
            "enhanced_head": enhanced_head_pred,
            "enhanced_deprel": enhanced_relations_pred,
            "sentence_embedding": torch.max(encoder_emb, dim=1)[0],
            "upostag_token_embedding": upos_output["embedding"],
            "xpostag_token_embedding": xpos_output["embedding"],
            "semrel_token_embedding": semrel_output["embedding"],
            "feats_token_embedding": morpho_output["embedding"],
            "deprel_token_embedding": parser_output["embedding"],
            "deprel_tree_distribution": parser_output["deprel_tree_distribution"],
            "deprel_label_distribution": parser_output["deprel_label_distribution"]
        }

        if "rel_probability" in enhanced_parser_output:
            output["enhanced_deprel_prob"] = enhanced_parser_output["rel_probability"]

        if self._has_labels([upostag, xpostag, lemma, feats, head, deprel, semrel]):

            # Feats mapping
            if self.morphological_feat:
                mapped_gold_labels = []
                for _, cat_indices in self.morphological_feat.slices.items():
                    try:
                        mapped_gold_labels.append(feats[:, :, cat_indices].argmax(dim=-1))
                    except TypeError:
                        raise ConfigurationError('Feats is None - if no feats are provided, the morphological_feat property should be set to None.')

                feats = torch.stack(mapped_gold_labels, dim=-1)

            labels = {
                "upostag": upostag,
                "xpostag": xpostag,
                "semrel": semrel,
                "feats": feats,
                "lemma": lemma.get("char").get("token_characters") if lemma else None,
                "head": head,
                "deprel": deprel,
                "enhanced_head": enhanced_heads,
                "enhanced_deprel": enhanced_deprels,
            }
            self.scores(output, labels, word_mask)
            relations_loss, head_loss = parser_output["loss"]
            enhanced_relations_loss, enhanced_head_loss = enhanced_parser_output["loss"]
            losses = {
                "upostag_loss": upos_output.get("loss"),
                "xpostag_loss": xpos_output.get("loss"),
                "semrel_loss": semrel_output.get("loss"),
                "feats_loss": morpho_output.get("loss"),
                "lemma_loss": lemma_output.get("loss"),
                "head_loss": head_loss,
                "deprel_loss": relations_loss,
                "enhanced_head_loss": enhanced_head_loss,
                "enhanced_deprel_loss": enhanced_relations_loss,
                # Cycle loss is only for the metrics purposes.
                "cycle_loss": parser_output.get("cycle_loss")
            }
            self._partial_losses = losses.copy()
            losses["loss"] = self._calculate_loss(losses)
            output.update(losses)

        return self._clean(output)

    def forward_on_instances(self, instances: List[Instance]) -> List[Dict[str, numpy.ndarray]]:
        batch_size = len(instances)
        with torch.no_grad():
            cuda_device = self._get_prediction_device()
            dataset = Batch(instances)
            dataset.index_instances(self.vocab)
            model_input = utils.move_to_device(dataset.as_tensor_dict(), cuda_device)
            outputs = self.make_output_human_readable(self.forward_on_tensors(**model_input))

            instance_separated_output: List[Dict[str, numpy.ndarray]] = [
                {} for _ in dataset.instances
            ]
            for name, output in list(outputs.items()):
                if isinstance(output, torch.Tensor):
                    # NOTE(markn): This is a hack because 0-dim pytorch tensors are not iterable.
                    # This occurs with batch size 1, because we still want to include the loss in that case.
                    if output.dim() == 0:
                        output = output.unsqueeze(0)

                    if output.size(0) != batch_size:
                        self._maybe_warn_for_unseparable_batches(name)
                        continue
                    output = output.detach().cpu().numpy()
                elif len(output) != batch_size:
                    self._maybe_warn_for_unseparable_batches(name)
                    continue
                for instance_output, batch_element in zip(instance_separated_output, output):
                    instance_output[name] = batch_element
            return instance_separated_output

    @staticmethod
    def _has_labels(labels):
        return any(x is not None for x in labels)

    def _calculate_loss(self, output):
        losses = []
        for name, value in self.loss_weights.items():
            if output.get(f"{name}_loss"):
                losses.append(output[f"{name}_loss"] * value)
        return torch.stack(losses).sum()

    @staticmethod
    def _optional(callable_model: Optional[torch.nn.Module],
                  *args,
                  returns_tuple: bool = False,
                  **kwargs):
        if callable_model:
            return callable_model(*args, **kwargs)
        if returns_tuple:
            return {"prediction": (None, None), "loss": (None, None), "embedding": (None, None)}
        return {"prediction": None, "loss": None, "embedding": None}

    @staticmethod
    def _clean(output):
        for k, v in dict(output).items():
            if v is None:
                del output[k]
        return output

    @overrides
    def get_metrics(self, reset: bool = False) -> Dict[str, float]:
        metrics = self.scores.get_metric(reset)
        if self._partial_losses:
            losses = self._clean(self._partial_losses)
            losses = {f"partial_loss/{k}": v.detach().item() for k, v in losses.items()}
            metrics.update(losses)
        return metrics

    @overrides
    def batch_outputs(self, batch: TensorDict, for_training: bool) -> Dict[str, torch.Tensor]:
        output_dict = self.forward_on_tensors(**batch)

        if for_training:
            try:
                assert "loss" in output_dict
                regularization_penalty = self.get_regularization_penalty()

                if regularization_penalty is not None:
                    output_dict["reg_loss"] = regularization_penalty
                    output_dict["loss"] += regularization_penalty

            except AssertionError:
                if for_training:
                    raise RuntimeError(
                        "The model you are trying to optimize does not contain a"
                        " 'loss' key in the output of model.forward(inputs)."
                    )

        return output_dict

    @overrides
    def forward(self, batch: TensorDict) -> TensorDict:
        return self.batch_outputs(batch, self.training)

    @overrides
    def training_step(self, batch: TensorDict, batch_idx: int) -> Tensor:
        output = self.forward(batch)
        self.log("train_loss", output['loss'], on_step=True, on_epoch=True, prog_bar=True, logger=True)
        return output["loss"]

    @overrides
    def validation_step(self, batch: TensorDict, batch_idx: int) -> Tensor:
        output = self.forward(batch)
        metrics = self.get_metrics()
        for k in metrics.keys():
            if k in self.validation_metrics:
                self.log(k, metrics[k], on_epoch=True, prog_bar=True, logger=True)
        return output["loss"]



import unittest

from combo.data import CharacterTokenizer


class CharacterTokenizerText(unittest.TestCase):

    def setUp(self) -> None:
        self.character_tokenizer = CharacterTokenizer()

    def test_tokenize_sentence(self):
        tokens = self.character_tokenizer.tokenize('I love you!')
        self.assertListEqual([t.text for t in tokens],
                             ['I', ' ', 'l', 'o', 'v', 'e', ' ', 'y', 'o', 'u', '!'])

    def test_tokenize_sentence_with_start_tokens(self):
        tokenizer_w_start_tokens = CharacterTokenizer(start_tokens=['@'])
        tokens = tokenizer_w_start_tokens.tokenize('Hi! Hello.')
        self.assertListEqual([t.text for t in tokens],
                             ['@', 'H', 'i', '!', ' ', 'H', 'e', 'l', 'l', 'o', '.'])
        self.assertEqual(tokens[0].idx, 0)
        self.assertTrue([t.idx > 0 for t in tokens if t.idx is not None])

    def test_tokenize_sentence_with_end_tokens(self):
        tokenizer_w_end_tokens = CharacterTokenizer(end_tokens=['#'])
        tokens = tokenizer_w_end_tokens.tokenize('Hi! Hello.')
        self.assertListEqual([t.text for t in tokens],
                             ['H', 'i', '!', ' ', 'H', 'e', 'l', 'l', 'o', '.', '#'])
        self.assertEqual(tokens[-1].idx, 0)
        self.assertTrue([t.idx > 0 for t in tokens if t.idx is not None])

    def test_tokenize_empty_sentence(self):
        tokens = self.character_tokenizer.tokenize('')
        self.assertEqual(len(tokens), 0)



import unittest

from combo.data import SpacyTokenizer


class SpacyTokenizerTest(unittest.TestCase):

    def setUp(self) -> None:
        self.spacy_tokenizer = SpacyTokenizer()

    def test_tokenize_sentence(self):
        tokens = self.spacy_tokenizer.tokenize('Hello cats. I love you')
        self.assertListEqual([t.text for t in tokens],
                             ['Hello', 'cats', '.', 'I', 'love', 'you'])

    def test_tokenize_empty_sentence(self):
        tokens = self.spacy_tokenizer.tokenize('')
        self.assertEqual(len(tokens), 0)




"""
Adapted from COMBO 1.0
Author: Mateusz Klimaszewski
"""

from typing import List

import torch

from combo.config import FromParameters, Registry
from combo.config.from_parameters import register_arguments
from combo.nn.activations import Activation


@Registry.register('dilated_cnn')
class DilatedCnnEncoder(torch.nn.Module, FromParameters):
    @register_arguments
    def __init__(self,
                 input_dim: int,
                 filters: List[int],
                 kernel_size: List[int],
                 stride: List[int],
                 padding: List[int],
                 dilation: List[int],
                 activations: List[Activation]):
        super().__init__()
        conv1d_layers = []
        input_dims = [input_dim] + filters[:-1]
        output_dims = filters
        for idx in range(len(activations)):
            conv1d_layers.append(torch.nn.Conv1d(
                in_channels=input_dims[idx],
                out_channels=output_dims[idx],
                kernel_size=(kernel_size[idx],),
                stride=(stride[idx],),
                padding=padding[idx],
                dilation=(dilation[idx],)))
        self.conv1d_layers = torch.nn.ModuleList(conv1d_layers)
        self.activations = activations
        assert len(self.activations) == len(self.conv1d_layers)

    def forward(self, x: torch.FloatTensor) -> torch.FloatTensor:
        for layer, activation in zip(self.conv1d_layers, self.activations):
            x = activation(layer(x))
        return x



from typing import Optional, Type, Dict, Tuple

from combo.config.exceptions import RegistryException


class Registry:
    __classes: Dict[str, Tuple[Type, str]] = {}
    __class_names: Dict[Tuple[Type[object], str], str] = {}

    @classmethod
    def classes(cls) -> Dict[str, Tuple[Type, str]]:
        return cls.__classes

    @classmethod
    def register(cls, registry_name: str, constructor_method: Optional[str] = None):
        def decorator(clz):
            nonlocal registry_name, constructor_method
            if constructor_method is None:
                constructor_method = '__init__'
            cls.__classes[registry_name.lower()] = (clz, constructor_method)
            cls.__class_names[(clz, constructor_method)] = registry_name.lower()
            return clz

        return decorator

    @classmethod
    def resolve(cls, class_name: str) -> Optional[Tuple[Type, str]]:
        try:
            return cls.__classes[class_name.lower()]
        except KeyError:
            raise RegistryException('No registered name ' + class_name.lower())

    @classmethod
    def get_class_name(cls, clz: Type[object], constructor_method: Optional[str] = None) -> Optional[str]:
        try:
            if constructor_method is None:
                constructor_method = '__init__'
            return cls.__class_names[(clz, constructor_method)]
        except KeyError:
            raise RegistryException('Class ' + str(clz) + ' is not registered!')



import re
from typing import List, Tuple, Dict, Any

import torch

from combo.config import FromParameters, Registry
from combo.config.from_parameters import register_arguments, resolve
from combo.nn.regularizers import Regularizer
from combo.utils.checks import ConfigurationError


@Registry.register('base_regularizer')
class RegularizerApplicator(FromParameters):
    """
    Applies regularizers to the parameters of a Module based on regex matches.
    """
    @register_arguments
    def __init__(self, regexes: List[Tuple[str, Regularizer]] = None) -> None:
        """
        # Parameters
        regexes : `List[Tuple[str, Regularizer]]`, optional (default = `None`)
            A sequence of pairs (regex, Regularizer), where each Regularizer
            applies to the parameters its regex matches (and that haven't previously
            been matched).
        """
        self._regularizers = regexes or []

    def __call__(self, module: torch.nn.Module) -> torch.Tensor:
        """
        # Parameters
        module : `torch.nn.Module`, required
            The module to regularize.
        """
        accumulator = 0.0
        for name, parameter in module.named_parameters():
            # We first check if the parameter needs gradient updates or not
            if parameter.requires_grad:
                # For each parameter find the first matching regex.
                for regex, regularizer in self._regularizers:
                    if re.search(regex, name):
                        penalty = regularizer(parameter)
                        accumulator = accumulator + penalty
                        break
        return accumulator

    @classmethod
    def from_parameters(cls,
                        parameters: Dict[str, Any] = None,
                        constructor_method_name: str = None,
                        pass_down_parameters: Dict[str, Any] = None):
        regexes = parameters.get('regexes', [])
        regexes_to_pass = []
        for regex, regularizer_dict in regexes:
            if isinstance(regularizer_dict, dict):
                if 'type' not in regularizer_dict:
                    raise ConfigurationError('Regularizer dict does not have the type field')
                resolved_regularizer = resolve(regularizer_dict, pass_down_parameters)
                regexes_to_pass.append((regex, resolved_regularizer))
            else:
                regexes_to_pass.append((regex, regularizer_dict))
        return cls(regexes_to_pass)



import unittest
import os

from combo.data.tokenizers import Token
from combo.data.token_indexers import PretrainedTransformerFixedMismatchedIndexer
from combo.data.vocabulary import Vocabulary


class TokenFeatsIndexerTest(unittest.TestCase):
    def setUp(self) -> None:
        self.indexer = PretrainedTransformerFixedMismatchedIndexer("allegro/herbert-base-cased")
        self.short_indexer = PretrainedTransformerFixedMismatchedIndexer("allegro/herbert-base-cased",
                                                                         max_length=3)
        self.vocabulary = Vocabulary.from_files(
            os.path.join(os.getcwd(), '../../fixtures/train_vocabulary'),
            oov_token='_',
            padding_token='__PAD__'
        )

    def test_offsets(self):
        output1 = self.indexer.tokens_to_indices([
            Token('Hello'), Token(','), Token('my'), Token('friend'), Token('!'),
            Token('What'), Token('a'), Token('nice'), Token('day'), Token('!')
        ], self.vocabulary)
        output2 = self.short_indexer.tokens_to_indices([
            Token('Hello'), Token(','), Token('my'), Token('friend'), Token('!'),
            Token('What'), Token('a'), Token('nice'), Token('day'), Token('!')
        ], self.vocabulary)
        self.assertListEqual(output1['offsets'], output2['offsets'])



"""Sampler tests."""
import unittest

from combo.data import TokenCountBatchSampler, Instance
from combo.data.fields.text_field import TextField
from combo.data.tokenizers import Token


class TokenCountBatchSamplerTest(unittest.TestCase):

    def setUp(self) -> None:
        self.dataset = []
        self.sentences = ["First sentence makes full batch.", "Short", "This ends first batch"]
        for sentence in self.sentences:
            tokens = [Token(t)
                      for t in sentence.split()]
            text_field = TextField(tokens, {})
            self.dataset.append(Instance({"sentence": text_field}))

    def test_batches(self):
        # given
        sampler = TokenCountBatchSampler(self.dataset, word_batch_size=2, shuffle_dataset=False)

        # when
        length = len(sampler)
        values = list(sampler)

        # then
        self.assertEqual(2, length)
        # sort by lengths + word_batch_size makes 1, 2 first batch
        self.assertListEqual([[1, 2], [0]], values)



import logging
import os
import sys
from typing import List, Union, Dict, Any

import numpy as np
import torch
from overrides import overrides

from combo import data
from combo.common import util
from combo.config import Registry
from combo.config.from_parameters import register_arguments
from combo.data import Instance, conllu2sentence, tokens2conllu, sentence2conllu
from combo.data.dataset_loaders.dataset_loader import TensorDict
from combo.data.dataset_readers.dataset_reader import DatasetReader
from combo.data.instance import JsonDict
from combo.default_model import default_ud_dataset_reader
from combo.modules.archival import load_archive
from combo.predictors import PredictorModule
from combo.utils import download, graph
from combo.modules.model import Model

logger = logging.getLogger(__name__)


@Registry.register('combo')
class COMBO(PredictorModule):
    @register_arguments
    def __init__(self,
                 model: Model,
                 dataset_reader: DatasetReader,
                 batch_size: int = 1024,
                 line_to_conllu: bool = True) -> None:
        super().__init__(model, dataset_reader)
        self.batch_size = batch_size
        self.vocab = model.vocab
        self.dataset_reader.generate_labels = False
        self.dataset_reader.lazy = True
        self.tokenizer = dataset_reader.tokenizer
        self.without_sentence_embedding = False
        self.line_to_conllu = line_to_conllu

    def __call__(self, sentence: Union[str, List[str], List[List[str]], List[data.Sentence]]):
        """Depending on the input uses (or ignores) tokenizer.
        When model isn't only text-based only List[data.Sentence] is possible input.

        * str - tokenizer is used
        * List[str] - tokenizer is used for each string (treated as list of raw sentences)
        * List[List[str]] - tokenizer isn't used (treated as list of tokenized sentences)
        * List[data.Sentence] - tokenizer isn't used (treated as list of tokenized sentences)

        :param sentence: sentence(s) representation
        :return: Sentence or List[Sentence] depending on the input
        """
        try:
            return self.predict(sentence)
        except Exception as e:
            logger.error(e)
            logger.error('Exiting.')
            sys.exit(1)

    def forward(self, inputs: TensorDict, training: bool = False) -> Dict[str, torch.Tensor]:
        return self.batch_outputs(inputs, training)

    def training_step(self, batch: TensorDict) -> Dict[str, torch.Tensor]:
        self._model.train()
        output = self.forward(batch)
        output['loss'].backward()
        return output

    def configure_optimizers(self) -> Any:
        pass

    def predict(self, sentence: Union[str, List[str], List[List[str]], List[data.Sentence]]):
        if isinstance(sentence, str):
            sentence = self.dataset_reader.tokenizer.segment(sentence)

        if isinstance(sentence, list):
            if len(sentence) == 0:
                return []
            example = sentence[0]
            sentences = sentence
            if isinstance(example, str) or isinstance(example, list):
                result = []
                sentences = [self._to_input_json(s) for s in sentences]
                for sentences_batch in util.lazy_groups_of(sentences, self.batch_size):
                    sentences_batch = self.predict_batch_json(sentences_batch)
                    result.extend(sentences_batch)
                return result
            elif isinstance(example, data.Sentence):
                result = []
                sentences = [self._to_input_instance(s) for s in sentences]
                for sentences_batch in util.lazy_groups_of(sentences, self.batch_size):
                    sentences_batch = self.predict_batch_instance(sentences_batch)
                    result.extend(sentences_batch)
                return result
            else:
                raise ValueError("List must have either sentences as str, List[str] or Sentence object.")
        else:
            raise ValueError("Input must be either string or list of strings.")

    @overrides
    def predict_batch_instance(self, instances: List[Instance]) -> List[data.Sentence]:
        sentences = []
        predictions = super().predict_batch_instance(instances)
        for prediction, instance in zip(predictions, instances):
            (tree, sentence_embedding, embeddings,
             relation_distribution, relation_label_distribution) = self._predictions_as_tree(prediction, instance)
            sentence = conllu2sentence(
                tree, sentence_embedding, embeddings,
                relation_distribution=relation_distribution,
                relation_label_distribution=relation_label_distribution
            )
            sentences.append(sentence)
        return sentences

    @overrides
    def predict_batch_json(self, inputs: List[JsonDict]) -> List[data.Sentence]:
        instances = self._batch_json_to_instances(inputs)
        sentences = self.predict_batch_instance(instances)
        return sentences

    @overrides
    def predict_instance(self, instance: Instance, serialize: bool = True) -> data.Sentence:
        predictions = super().predict_instance(instance)
        (tree, sentence_embedding, embeddings,
         relation_distribution, relation_label_distribution) = self._predictions_as_tree(predictions, instance)
        return conllu2sentence(
            tree, sentence_embedding, embeddings,
            relation_distribution=relation_distribution,
            relation_label_distribution=relation_label_distribution
        )

    @overrides
    def predict_json(self, inputs: JsonDict) -> data.Sentence:
        instance = self._json_to_instance(inputs)
        return self.predict_instance(instance)

    @overrides
    def _json_to_instance(self, json_dict) -> Instance:
        sentence = json_dict["sentence"]
        if isinstance(sentence, str):
            tokens = [t.text for t in self.tokenizer.tokenize(json_dict["sentence"])]
        elif isinstance(sentence, list):
            tokens = sentence
        else:
            raise ValueError("Input must be either string or list of strings.")
        return self.dataset_reader.text_to_instance(tokens2conllu(tokens))

    @overrides
    def load_line(self, line: str) -> JsonDict:
        return self._to_input_json(line.replace("\n", "").strip())

    # outputs should be api.Sentence
    @overrides
    def dump_line(self, outputs: Any) -> str:
        # Check whether serialized (str) tree or token's list
        # Serialized tree has already separators between lines
        if self.without_sentence_embedding:
            outputs.sentence_embedding = []
        if self.line_to_conllu:
            return sentence2conllu(outputs, keep_semrel=getattr(self.dataset_reader, 'use_sem', False)).serialize()
        else:
            return outputs.to_json()

    @staticmethod
    def _to_input_json(sentence: str):
        return {"sentence": sentence}

    def _to_input_instance(self, sentence: data.Sentence) -> Instance:
        return self.dataset_reader.text_to_instance(sentence2conllu(sentence))

    def _predictions_as_tree(self, predictions: Dict[str, Any], instance: Instance):
        tree = instance.fields["metadata"]["input"]
        field_names = instance.fields["metadata"]["field_names"]
        tree_tokens = [t for t in tree if isinstance(t["idx"], int)]
        embeddings = {t["idx"]: {} for t in tree}
        deprel_tree_distribution = None
        deprel_label_distribution = None
        for field_name in field_names:
            if field_name not in predictions:
                continue
            if field_name == "deprel":
                sentence_length = len(tree_tokens)
                deprel_tree_distribution = np.matrix(predictions["deprel_tree_distribution"])[:sentence_length + 1,
                                           :sentence_length + 1]
                deprel_label_distribution = np.matrix(predictions["deprel_label_distribution"])[:sentence_length, :]
            field_predictions = predictions[field_name]
            for idx, token in enumerate(tree_tokens):
                if field_name in {"xpostag", "upostag", "semrel", "deprel"}:
                    value = self.vocab.get_token_from_index(field_predictions[idx], field_name + "_labels")
                    setattr(token, field_name, value)
                    embeddings[token["idx"]][field_name] = predictions[f"{field_name}_token_embedding"][idx]
                elif field_name == "head":
                    setattr(token, field_name, int(field_predictions[idx]))
                elif field_name == "deps":
                    # Handled after every other decoding
                    continue

                elif field_name == "feats":
                    slices = self._model.morphological_feat.slices
                    features = []
                    prediction = field_predictions[idx]
                    for (cat, cat_indices), pred_idx in zip(slices.items(), prediction):
                        if cat not in ["__PAD__", "_"]:
                            value = self.vocab.get_token_from_index(cat_indices[pred_idx],
                                                                    field_name + "_labels")
                            # Exclude auxiliary values
                            if "=None" not in value:
                                features.append(value)
                    if len(features) == 0:
                        field_value = "_"
                    else:
                        lowercase_features = [f.lower() for f in features]
                        arg_indices = sorted(range(len(lowercase_features)), key=lowercase_features.__getitem__)
                        field_value = "|".join(np.array(features)[arg_indices].tolist())

                    setattr(token, field_name, field_value)
                    embeddings[token["idx"]][field_name] = predictions[f"{field_name}_token_embedding"][idx]
                elif field_name == "lemma":
                    prediction = field_predictions[idx]
                    word_chars = []
                    for char_idx in prediction[1:-1]:
                        pred_char = self.vocab.get_token_from_index(char_idx, "lemma_characters")

                        if pred_char == "__END__":
                            break
                        elif pred_char == "__PAD__":
                            continue
                        elif "_" in pred_char:
                            pred_char = "?"

                        word_chars.append(pred_char)
                    setattr(token, field_name, "".join(word_chars))
                else:
                    raise NotImplementedError(f"Unknown field name {field_name}!")

        if "enhanced_head" in predictions and predictions["enhanced_head"]:
            # TODO off-by-one hotfix, refactor
            sentence_length = len(tree_tokens)
            h = np.array(predictions["enhanced_head"])[:sentence_length, :sentence_length]
            h = np.concatenate((h[-1:], h[:-1]))
            r = np.array(predictions["enhanced_deprel_prob"])[:sentence_length, :sentence_length, :]
            r = np.concatenate((r[-1:], r[:-1]))

            graph.graph_and_tree_merge(
                tree_arc_scores=predictions["head"][:sentence_length],
                tree_rel_scores=predictions["deprel"][:sentence_length],
                graph_arc_scores=h,
                graph_rel_scores=r,
                idx2label=self.vocab.get_index_to_token_vocabulary("deprel_labels"),
                label2idx=self.vocab.get_token_to_index_vocabulary("deprel_labels"),
                graph_idx2label=self.vocab.get_index_to_token_vocabulary("enhanced_deprel_labels"),
                graph_label2idx=self.vocab.get_token_to_index_vocabulary("enhanced_deprel_labels"),
                tokens=tree_tokens
            )

            empty_tokens = graph.restore_collapse_edges(tree_tokens)
            tree.tokens.extend(empty_tokens)

        return tree, predictions["sentence_embedding"], embeddings, \
               deprel_tree_distribution, deprel_label_distribution

    @classmethod
    def from_pretrained(cls,
                        path: str,
                        batch_size: int = 1024,
                        cuda_device: int = -1):
        if os.path.exists(path):
            model_path = path
        else:
            try:
                logger.debug("Downloading model.")
                model_path = download.download_file(path)
            except Exception as e:
                logger.error(e)
                raise e

        archive = load_archive(model_path, cuda_device=cuda_device)
        model = archive.model
        dataset_reader = archive.dataset_reader or default_ud_dataset_reader(archive.config.get("model_name"))
        return cls(model, dataset_reader, batch_size)



"""
Adapted from COMBO
https://gitlab.clarin-pl.eu/syntactic-tools/combo/-/blob/master/combo/models/base.py
"""

import torch
from combo.config.from_parameters import FromParameters
from typing import Dict, List, Optional, Union


class Predictor(torch.nn.Module, FromParameters):
    def forward(self,
                x: Union[torch.Tensor, List[torch.Tensor]],
                mask: Optional[torch.BoolTensor] = None,
                labels: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None,
                sample_weights: Optional[Union[torch.Tensor, List[torch.Tensor]]] = None) -> Dict[str, torch.Tensor]:
        raise NotImplementedError()



"""
Sequence multilabel field tests.
Adapted from original COMBO
Author: Mateusz Klimaszewski
"""

import unittest
from typing import List

import torch

from combo.data.fields.sequence_field import SequenceField
from combo.utils import pad_sequence_to_length

from combo.data import fields, Vocabulary


class IndexingSequenceMultiLabelFieldTest(unittest.TestCase):

    def setUp(self) -> None:
        self.namespace = "test_labels"
        self.vocab = Vocabulary()
        self.vocab.add_tokens_to_namespace(
            tokens=["t" + str(idx) for idx in range(3)],
            namespace=self.namespace
        )

        def _indexer(vocab: Vocabulary):
            vocab_size = vocab.get_vocab_size(self.namespace)

            def _mapper(multi_label: List[str], _: int) -> List[int]:
                one_hot = [0] * vocab_size
                for label in multi_label:
                    index = vocab.get_token_index(label, self.namespace)
                    one_hot[index] = 1
                return one_hot

            return _mapper

        def _as_tensor(field: fields.SequenceMultiLabelField):
            def _wrapped(padding_lengths):
                desired_num_tokens = padding_lengths["num_tokens"]
                classes_count = len(field._indexed_multi_labels[0])
                default_value = [0.0] * classes_count
                padded_tags = pad_sequence_to_length(field._indexed_multi_labels, desired_num_tokens,
                                                     lambda: default_value)
                tensor = torch.LongTensor(padded_tags)
                return tensor

            return _wrapped

        self.indexer = _indexer
        self.as_tensor = _as_tensor
        self.sequence_field = _SequenceFieldTestWrapper(self.vocab.get_vocab_size(self.namespace))

    def test_indexing(self):
        # given
        field = fields.SequenceMultiLabelField(
            multi_labels=[["t1", "t2"], [], ["t0"]],
            multi_label_indexer=self.indexer,
            as_tensor=self.as_tensor,
            sequence_field=self.sequence_field,
            label_namespace=self.namespace
        )
        expected = [[0, 1, 1], [0, 0, 0], [1, 0, 0]]

        # when
        field.index(self.vocab)

        # then
        self.assertEqual(expected, field._indexed_multi_labels)

    def test_mapping_to_tensor(self):
        # given
        field = fields.SequenceMultiLabelField(
            multi_labels=[["t1", "t2"], [], ["t0"]],
            multi_label_indexer=self.indexer,
            as_tensor=self.as_tensor,
            sequence_field=self.sequence_field,
            label_namespace=self.namespace
        )
        field.index(self.vocab)
        expected = torch.tensor([[0, 1, 1], [0, 0, 0], [1, 0, 0]])

        # when
        actual = field.as_tensor(field.get_padding_lengths())

        # then
        self.assertTrue(torch.all(expected.eq(actual)))

    def test_sequence_method(self):
        # given
        field = fields.SequenceMultiLabelField(
            multi_labels=[["t1", "t2"], [], ["t0"]],
            multi_label_indexer=self.indexer,
            as_tensor=self.as_tensor,
            sequence_field=self.sequence_field,
            label_namespace=self.namespace
        )

        # when
        length = len(field)
        iter_length = len(list(iter(field)))
        middle_value = field[1]

        # then
        self.assertEqual(3, length)
        self.assertEqual(3, iter_length)
        self.assertEqual([], middle_value)


class _SequenceFieldTestWrapper(SequenceField):

    def __init__(self, length: int):
        self.length = length

    def sequence_length(self) -> int:
        return self.length



from typing import Optional, NamedTuple

import h5py
import torch
from overrides import overrides
from torchtext.vocab import Vectors, GloVe, FastText, CharNGram

from combo import nn
from combo.config import FromParameters, Registry
from combo.config.from_parameters import register_arguments
from combo.data import Vocabulary
from combo.nn.utils import tiny_value_of_dtype, uncombine_initial_dims, combine_initial_dims
from combo.modules.module import Module
from combo.utils import ConfigurationError
from combo.models.time_distributed import TimeDistributed


class TokenEmbedder(Module, FromParameters):
    def __init__(self):
        super(TokenEmbedder, self).__init__()

    def get_output_dim(self) -> int:
        raise NotImplementedError()

    def forward(self,
                x: torch.Tensor,
                char_mask: Optional[torch.BoolTensor] = None) -> torch.Tensor:
        raise NotImplementedError()


@Registry.register("embedding")
class Embedding(TokenEmbedder):
    """
    A more featureful embedding module than the default in Pytorch.  Adds the ability to:

        1. embed higher-order inputs
        2. pre-specify the weight matrix
        3. use a non-trainable embedding
        4. project the resultant embeddings to some other dimension (which only makes sense with
           non-trainable embeddings).

    Note that if you are using our data API and are trying to embed a
    [`TextField`](../../data/fields/text_field.md), you should use a
    [`TextFieldEmbedder`](../text_field_embedders/text_field_embedder.md) instead of using this directly.

    Registered as a `TokenEmbedder` with name "embedding".

    # Parameters

    num_embeddings : `int`
        Size of the dictionary of embeddings (vocabulary size).
    embedding_dim : `int`
        The size of each embedding vector.
    projection_dim : `int`, optional (default=`None`)
        If given, we add a projection layer after the embedding layer.  This really only makes
        sense if `trainable` is `False`.
    weight : `torch.FloatTensor`, optional (default=`None`)
        A pre-initialised weight matrix for the embedding lookup, allowing the use of
        pretrained vectors.
    padding_index : `int`, optional (default=`None`)
        If given, pads the output with zeros whenever it encounters the index.
    trainable : `bool`, optional (default=`True`)
        Whether or not to optimize the embedding parameters.
    max_norm : `float`, optional (default=`None`)
        If given, will renormalize the embeddings to always have a norm lesser than this
    norm_type : `float`, optional (default=`2`)
        The p of the p-norm to compute for the max_norm option
    scale_grad_by_freq : `bool`, optional (default=`False`)
        If given, this will scale gradients by the frequency of the words in the mini-batch.
    sparse : `bool`, optional (default=`False`)
        Whether or not the Pytorch backend should use a sparse representation of the embedding weight.
    vocab_namespace : `str`, optional (default=`None`)
        In case of fine-tuning/transfer learning, the model's embedding matrix needs to be
        extended according to the size of extended-vocabulary. To be able to know how much to
        extend the embedding-matrix, it's necessary to know which vocab_namspace was used to
        construct it in the original training. We store vocab_namespace used during the original
        training as an attribute, so that it can be retrieved during fine-tuning.
    pretrained_file : `str`, optional (default=`None`)
        Path to a file of word vectors to initialize the embedding matrix. It can be the
        path to a local file or a URL of a (cached) remote file. Two formats are supported:
            * hdf5 file - containing an embedding matrix in the form of a torch.Tensor;
            * text file - an utf-8 encoded text file with space separated fields.
    vocabulary : `Vocabulary`, optional (default = `None`)
        Used to construct an embedding from a pretrained file.

        In a typical AllenNLP configuration file, this parameter does not get an entry under the
        "embedding", it gets specified as a top-level parameter, then is passed in to this module
        separately.

    # Returns

    An Embedding module.
    """

    @register_arguments
    def __init__(
            self,
            embedding_dim: int,
            num_embeddings: int = None,
            projection_dim: int = None,
            weight: torch.FloatTensor = None,
            padding_index: int = None,
            trainable: bool = True,
            max_norm: float = None,
            norm_type: float = 2.0,
            scale_grad_by_freq: bool = False,
            sparse: bool = False,
            vocab_namespace: str = "tokens",
            pretrained_file: str = None,
            vocabulary: Vocabulary = None,
    ) -> None:
        super().__init__()

        if num_embeddings is None and vocabulary is None:
            raise ConfigurationError(
                "Embedding must be constructed with either num_embeddings or a vocabulary."
            )

        _vocab_namespace: Optional[str] = vocab_namespace
        if num_embeddings is None:
            num_embeddings = vocabulary.get_vocab_size(_vocab_namespace)  # type: ignore
        else:
            # If num_embeddings is present, set default namespace to None so that extend_vocab
            # call doesn't misinterpret that some namespace was originally used.
            _vocab_namespace = None  # type: ignore

        self.num_embeddings = num_embeddings
        self.padding_index = padding_index
        self.max_norm = max_norm
        self.norm_type = norm_type
        self.scale_grad_by_freq = scale_grad_by_freq
        self.sparse = sparse
        self._vocab_namespace = _vocab_namespace
        self._pretrained_file = pretrained_file

        self.output_dim = projection_dim or embedding_dim

        if weight is not None and pretrained_file:
            raise ConfigurationError(
                "Embedding was constructed with both a weight and a pretrained file."
            )

        elif pretrained_file is not None:

            if vocabulary is None:
                raise ConfigurationError(
                    "To construct an Embedding from a pretrained file, you must also pass a vocabulary."
                )

            # If we're loading a saved model, we don't want to actually read a pre-trained
            # embedding file - the embeddings will just be in our saved weights, and we might not
            # have the original embedding file anymore, anyway.

            # TODO: having to pass tokens here is SUPER gross, but otherwise this breaks the
            # extend_vocab method, which relies on the value of vocab_namespace being None
            # to infer at what stage the embedding has been constructed. Phew.
            weight = _read_pretrained_embeddings_file(
                pretrained_file, embedding_dim, vocabulary, vocab_namespace
            )
            self.weight = torch.nn.Parameter(weight, requires_grad=trainable)

        elif weight is not None:
            self.weight = torch.nn.Parameter(weight, requires_grad=trainable)

        else:
            weight = torch.FloatTensor(num_embeddings, embedding_dim)
            self.weight = torch.nn.Parameter(weight, requires_grad=trainable)
            torch.nn.init.xavier_uniform_(self.weight)

        # Whatever way we have constructed the embedding, it should be consistent with
        # num_embeddings and embedding_dim.
        if self.weight.size() != (num_embeddings, embedding_dim):
            raise ConfigurationError(
                "A weight matrix was passed with contradictory embedding shapes."
            )

        if self.padding_index is not None:
            self.weight.data[self.padding_index].fill_(0)

        if projection_dim:
            self._projection = torch.nn.Linear(embedding_dim, projection_dim)
        else:
            self._projection = None

    def get_output_dim(self) -> int:
        return self.output_dim

    def forward(self, tokens: torch.Tensor) -> torch.Tensor:
        # tokens may have extra dimensions (batch_size, d1, ..., dn, sequence_length),
        # but embedding expects (batch_size, sequence_length), so pass tokens to
        # util.combine_initial_dims (which is a no-op if there are no extra dimensions).
        # Remember the original size.
        original_size = tokens.size()
        tokens = combine_initial_dims(tokens)

        embedded = embedding(
            tokens,
            self.weight,
            padding_idx=self.padding_index,
            max_norm=self.max_norm,
            norm_type=self.norm_type,
            scale_grad_by_freq=self.scale_grad_by_freq,
            sparse=self.sparse,
        )

        # Now (if necessary) add back in the extra dimensions.
        embedded = uncombine_initial_dims(embedded, original_size)

        if self._projection:
            projection = self._projection
            for _ in range(embedded.dim() - 2):
                projection = TimeDistributed(projection)
            embedded = projection(embedded)
        return embedded

    def extend_vocab(
            self,
            extended_vocab: Vocabulary,
            vocab_namespace: str = None,
            extension_pretrained_file: str = None,
            model_path: str = None,
    ):
        """
        Extends the embedding matrix according to the extended vocabulary.
        If extension_pretrained_file is available, it will be used for initializing the new words
        embeddings in the extended vocabulary; otherwise we will check if _pretrained_file attribute
        is already available. If none is available, they will be initialized with xavier uniform.

        # Parameters

        extended_vocab : `Vocabulary`
            Vocabulary extended from original vocabulary used to construct
            this `Embedding`.
        vocab_namespace : `str`, (optional, default=`None`)
            In case you know what vocab_namespace should be used for extension, you
            can pass it. If not passed, it will check if vocab_namespace used at the
            time of `Embedding` construction is available. If so, this namespace
            will be used or else extend_vocab will be a no-op.
        extension_pretrained_file : `str`, (optional, default=`None`)
            A file containing pretrained embeddings can be specified here. It can be
            the path to a local file or an URL of a (cached) remote file. Check format
            details in `from_params` of `Embedding` class.
        model_path : `str`, (optional, default=`None`)
            Path traversing the model attributes upto this embedding module.
            Eg. "_text_field_embedder.token_embedder_tokens". This is only useful
            to give a helpful error message when extend_vocab is implicitly called
            by train or any other command.
        """
        # Caveat: For allennlp v0.8.1 and below, we weren't storing vocab_namespace as an attribute,
        # knowing which is necessary at time of embedding vocabulary extension. So old archive models are
        # currently unextendable.

        vocab_namespace = vocab_namespace or self._vocab_namespace
        if not vocab_namespace:
            # It's not safe to default to "tokens" or any other namespace.
            logger.info(
                "Loading a model trained before embedding extension was implemented; "
                "pass an explicit vocabulary namespace if you want to extend the vocabulary."
            )
            return

        extended_num_embeddings = extended_vocab.get_vocab_size(vocab_namespace)
        if extended_num_embeddings == self.num_embeddings:
            # It's already been extended. No need to initialize / read pretrained file in first place (no-op)
            return

        if extended_num_embeddings < self.num_embeddings:
            raise ConfigurationError(
                f"Size of namespace, {vocab_namespace} for extended_vocab is smaller than "
                f"embedding. You likely passed incorrect vocabulary or namespace for extension."
            )

        # Case 1: user passed extension_pretrained_file and it's available.
        if extension_pretrained_file and is_url_or_existing_file(extension_pretrained_file):
            # Don't have to do anything here, this is the happy case.
            pass
        # Case 2: user passed extension_pretrained_file and it's not available
        elif extension_pretrained_file:
            raise ConfigurationError(
                f"You passed pretrained embedding file {extension_pretrained_file} "
                f"for model_path {model_path} but it's not available."
            )
        # Case 3: user didn't pass extension_pretrained_file, but pretrained_file attribute was
        # saved during training and is available.
        elif is_url_or_existing_file(self._pretrained_file):
            extension_pretrained_file = self._pretrained_file
        # Case 4: no file is available, hope that pretrained embeddings weren't used in the first place and warn
        elif self._pretrained_file is not None:
            # Warn here instead of an exception to allow a fine-tuning even without the original pretrained_file
            logger.warning(
                f"Embedding at model_path, {model_path} cannot locate the pretrained_file. "
                f"Originally pretrained_file was at '{self._pretrained_file}'."
            )
        else:
            # When loading a model from archive there is no way to distinguish between whether a pretrained-file
            # was or wasn't used during the original training. So we leave an info here.
            logger.info(
                "If you are fine-tuning and want to use a pretrained_file for "
                "embedding extension, please pass the mapping by --embedding-sources argument."
            )

        embedding_dim = self.weight.data.shape[-1]
        if not extension_pretrained_file:
            extra_num_embeddings = extended_num_embeddings - self.num_embeddings
            extra_weight = torch.FloatTensor(extra_num_embeddings, embedding_dim)
            torch.nn.init.xavier_uniform_(extra_weight)
        else:
            # It's easiest to just reload the embeddings for the entire vocabulary,
            # then only keep the ones we need.
            whole_weight = _read_pretrained_embeddings_file(
                extension_pretrained_file, embedding_dim, extended_vocab, vocab_namespace
            )
            extra_weight = whole_weight[self.num_embeddings:, :]

        device = self.weight.data.device
        extended_weight = torch.cat([self.weight.data, extra_weight.to(device)], dim=0)
        self.weight = torch.nn.Parameter(extended_weight, requires_grad=self.weight.requires_grad)
        self.num_embeddings = extended_num_embeddings


def _read_pretrained_embeddings_file(
        file_uri: str, embedding_dim: int, vocab: Vocabulary, namespace: str = "tokens"
) -> torch.FloatTensor:
    """
    Returns and embedding matrix for the given vocabulary using the pretrained embeddings
    contained in the given file. Embeddings for tokens not found in the pretrained embedding file
    are randomly initialized using a normal distribution with mean and standard deviation equal to
    those of the pretrained embeddings.

    We support two file formats:

        * text format - utf-8 encoded text file with space separated fields: [word] [dim 1] [dim 2] ...
          The text file can eventually be compressed, and even resides in an archive with multiple files.
          If the file resides in an archive with other files, then `embeddings_filename` must
          be a URI "(archive_uri)#file_path_inside_the_archive"

        * hdf5 format - hdf5 file containing an embedding matrix in the form of a torch.Tensor.

    If the filename ends with '.hdf5' or '.h5' then we load from hdf5, otherwise we assume
    text format.

    # Parameters

    file_uri : `str`, required.
        It can be:

        * a file system path or a URL of an eventually compressed text file or a zip/tar archive
          containing a single file.

        * URI of the type `(archive_path_or_url)#file_path_inside_archive` if the text file
          is contained in a multi-file archive.

    vocabulary : `Vocabulary`, required.
        A Vocabulary object.
    namespace : `str`, (optional, default=`"tokens"`)
        The namespace of the vocabulary to find pretrained embeddings for.
    trainable : `bool`, (optional, default=`True`)
        Whether or not the embedding parameters should be optimized.

    # Returns

    A weight matrix with embeddings initialized from the read file.  The matrix has shape
    `(vocabulary.get_vocab_size(namespace), embedding_dim)`, where the indices of words appearing in
    the pretrained embedding file are initialized to the pretrained embedding value.
    """
    file_ext = get_file_extension(file_uri)
    if file_ext in [".h5", ".hdf5"]:
        return _read_embeddings_from_hdf5(file_uri, embedding_dim, vocab, namespace)

    return _read_embeddings_from_text_file(file_uri, embedding_dim, vocab, namespace)


def _read_embeddings_from_text_file(
        file_uri: str, embedding_dim: int, vocab: Vocabulary, namespace: str = "tokens"
) -> torch.FloatTensor:
    """
    Read pre-trained word vectors from an eventually compressed text file, possibly contained
    inside an archive with multiple files. The text file is assumed to be utf-8 encoded with
    space-separated fields: [word] [dim 1] [dim 2] ...

    Lines that contain more numerical tokens than `embedding_dim` raise a warning and are skipped.

    The remainder of the docstring is identical to `_read_pretrained_embeddings_file`.
    """
    tokens_to_keep = set(vocab.get_index_to_token_vocabulary(namespace).values())
    vocab_size = vocab.get_vocab_size(namespace)
    embeddings = {}

    # First we read the embeddings from the file, only keeping vectors for the words we need.
    logger.info("Reading pretrained embeddings from file")

    with EmbeddingsTextFile(file_uri) as embeddings_file:
        for line in Tqdm.tqdm(embeddings_file):
            token = line.split(" ", 1)[0]
            if token in tokens_to_keep:
                fields = line.rstrip().split(" ")
                if len(fields) - 1 != embedding_dim:
                    # Sometimes there are funny unicode parsing problems that lead to different
                    # fields lengths (e.g., a word with a unicode space character that splits
                    # into more than one column).  We skip those lines.  Note that if you have
                    # some kind of long header, this could result in all of your lines getting
                    # skipped.  It's hard to check for that here; you just have to look in the
                    # embedding_misses_file and at the model summary to make sure things look
                    # like they are supposed to.
                    logger.warning(
                        "Found line with wrong number of dimensions (expected: %d; actual: %d): %s",
                        embedding_dim,
                        len(fields) - 1,
                        line,
                    )
                    continue

                vector = numpy.asarray(fields[1:], dtype="float32")
                embeddings[token] = vector

    if not embeddings:
        raise ConfigurationError(
            "No embeddings of correct dimension found; you probably "
            "misspecified your embedding_dim parameter, or didn't "
            "pre-populate your Vocabulary"
        )

    all_embeddings = numpy.asarray(list(embeddings.values()))
    embeddings_mean = float(numpy.mean(all_embeddings))
    embeddings_std = float(numpy.std(all_embeddings))
    # Now we initialize the weight matrix for an embedding layer, starting with random vectors,
    # then filling in the word vectors we just read.
    logger.info("Initializing pre-trained embedding layer")
    embedding_matrix = torch.FloatTensor(vocab_size, embedding_dim).normal_(
        embeddings_mean, embeddings_std
    )
    num_tokens_found = 0
    index_to_token = vocab.get_index_to_token_vocabulary(namespace)
    for i in range(vocab_size):
        token = index_to_token[i]

        # If we don't have a pre-trained vector for this word, we'll just leave this row alone,
        # so the word has a random initialization.
        if token in embeddings:
            embedding_matrix[i] = torch.FloatTensor(embeddings[token])
            num_tokens_found += 1
        else:
            logger.debug(
                "Token %s was not found in the embedding file. Initialising randomly.", token
            )

    logger.info(
        "Pretrained embeddings were found for %d out of %d tokens", num_tokens_found, vocab_size
    )

    return embedding_matrix


def _read_embeddings_from_hdf5(
        embeddings_filename: str, embedding_dim: int, vocabulary: Vocabulary, namespace: str = "tokens"
) -> torch.FloatTensor:
    """
    Reads from a hdf5 formatted file. The embedding matrix is assumed to
    be keyed by 'embedding' and of size `(num_tokens, embedding_dim)`.
    """
    with h5py.File(embeddings_filename, "r") as fin:
        embeddings = fin["embedding"][...]

    if list(embeddings.shape) != [vocabulary.get_vocab_size(namespace), embedding_dim]:
        raise ConfigurationError(
            "Read shape {0} embeddings from the file, but expected {1}".format(
                list(embeddings.shape), [vocabulary.get_vocab_size(namespace), embedding_dim]
            )
        )

    return torch.FloatTensor(embeddings)


def format_embeddings_file_uri(
        main_file_path_or_url: str, path_inside_archive: Optional[str] = None
) -> str:
    if path_inside_archive:
        return "({})#{}".format(main_file_path_or_url, path_inside_archive)
    return main_file_path_or_url


class EmbeddingsFileURI(NamedTuple):
    main_file_uri: str
    path_inside_archive: Optional[str] = None


@Registry.register('base_token_embedder')
class TorchEmbedder(TokenEmbedder):
    @register_arguments
    def __init__(self,
                 num_TokenEmbedders: int,
                 TokenEmbedder_dim: int,
                 padding_idx: Optional[int] = None,
                 max_norm: Optional[float] = None,
                 norm_type: float = 2.,
                 scale_grad_by_freq: bool = False,
                 sparse: bool = False,
                 vocab_namespace: str = "tokens",
                 vocabulary: Vocabulary = None,
                 weight: Optional[torch.Tensor] = None,
                 trainable: bool = True,
                 projection_dim: Optional[int] = None):
        super(TorchEmbedder, self).__init__()
        self._TokenEmbedder_dim = TokenEmbedder_dim
        self._TokenEmbedder = nn.TokenEmbedder(num_TokenEmbedders=num_TokenEmbedders,
                                               TokenEmbedder_dim=TokenEmbedder_dim,
                                               padding_idx=padding_idx,
                                               max_norm=max_norm,
                                               norm_type=norm_type,
                                               scale_grad_by_freq=scale_grad_by_freq,
                                               sparse=sparse)
        self.__vocab_namespace = vocab_namespace
        self.__vocab = vocabulary

        if weight is not None:
            if weight.shape() != (num_TokenEmbedders, TokenEmbedder_dim):
                raise ConfigurationError(
                    "Weight matrix must be of shape (num_TokenEmbedders, TokenEmbedder_dim)." +
                    f"Got: ({weight.shape()})"
                )

            self.__weight = torch.nn.Parameter(weight, requires_grad=trainable)
        else:
            self.__weight = torch.nn.Parameter(torch.FloatTensor(num_TokenEmbedders, TokenEmbedder_dim),
                                               requires_grad=trainable)
            torch.nn.init.xavier_uniform_(self.__weight)

        if padding_idx is not None:
            self.__weight.data[padding_idx].fill_(0)

        if projection_dim:
            self._projection = torch.nn.Linear(TokenEmbedder_dim, projection_dim)
            self.output_dim = projection_dim
        else:
            self._projection = None
            self.output_dim = TokenEmbedder_dim

    @overrides
    def forward(self,
                x: torch.Tensor,
                char_mask: Optional[torch.BoolTensor] = None) -> torch.Tensor:
        embedded = self._TokenEmbedder(x)
        if self._projection:
            projection = self._projection
            for p in range(embedded.dim() - 2):
                projection = TimeDistributed(p)
            embedded = projection(embedded)
        return embedded


@Registry.register('torchtext_vectors_embedder')
class _TorchtextVectorsEmbedder(TokenEmbedder):
    """
    Torchtext Vectors object wrapper
    """

    @register_arguments
    def __init__(self,
                 torchtext_embedder: Vectors,
                 lower_case_backup: bool = False):
        """
        :param torchtext_embedder: Torchtext Vectors object
        :param lower_case_backup: whether to look up the token in the
        lower case. Default: False.
        """
        super(_TorchtextVectorsEmbedder, self).__init__()
        self.__torchtext_embedder = torchtext_embedder
        self.__lower_case_backup = lower_case_backup

    @overrides
    def get_output_dim(self) -> int:
        return len(self.__torchtext_embedder)

    @overrides
    def forward(self,
                x: torch.Tensor,
                char_mask: Optional[torch.BoolTensor] = None) -> torch.Tensor:
        return self.__torchtext_embedder.get_vecs_by_tokens(x, self.__lower_case_backup)


@Registry.register('glove42b')
class GloVe42BEmbedder(_TorchtextVectorsEmbedder):
    @register_arguments
    def __init__(self, dim: int = 300):
        super(GloVe42BEmbedder, self).__init__(GloVe("42B", dim))


@Registry.register('glove840b')
class GloVe840BEmbedder(_TorchtextVectorsEmbedder):
    @register_arguments
    def __init__(self, dim: int = 300):
        super(GloVe840BEmbedder, self).__init__(GloVe("840B", dim))


@Registry.register('glove_twitter27b')
class GloVeTwitter27BEmbedder(_TorchtextVectorsEmbedder):
    @register_arguments
    def __init__(self, dim: int = 300):
        super(GloVeTwitter27BEmbedder, self).__init__(GloVe("twitter.27B", dim))


@Registry.register('glove6b')
class GloVe6BEmbedder(_TorchtextVectorsEmbedder):
    @register_arguments
    def __init__(self, dim: int = 300):
        super(GloVe6BEmbedder, self).__init__(GloVe("6B", dim))


@Registry.register('fast_text')
class FastTextEmbedder(_TorchtextVectorsEmbedder):
    @register_arguments
    def __init__(self, language: str = "en"):
        super(FastTextEmbedder, self).__init__(FastText(language))


@Registry.register('char_ngram')
class CharNGramEmbedder(_TorchtextVectorsEmbedder):
    @register_arguments
    def __init__(self):
        super(CharNGramEmbedder, self).__init__(CharNGram())


@Registry.register('feats_token_embedder')
class FeatsTokenEmbedder(TorchEmbedder):
    @register_arguments
    def __init__(self,
                 num_TokenEmbedders: int,
                 TokenEmbedder_dim: int,
                 padding_idx: Optional[int] = None,
                 max_norm: Optional[float] = None,
                 norm_type: float = 2.,
                 scale_grad_by_freq: bool = False,
                 sparse: bool = False,
                 vocab_namespace: str = "feats",
                 vocabulary: Vocabulary = None,
                 weight: Optional[torch.Tensor] = None,
                 trainable: bool = True):
        super(FeatsTokenEmbedder, self).__init__(num_TokenEmbedders,
                                                 TokenEmbedder_dim,
                                                 padding_idx,
                                                 max_norm,
                                                 norm_type,
                                                 scale_grad_by_freq,
                                                 sparse,
                                                 vocab_namespace,
                                                 vocabulary,
                                                 weight,
                                                 trainable)

    @overrides
    def forward(self,
                x: torch.Tensor,
                char_mask: Optional[torch.BoolTensor] = None) -> torch.Tensor:
        mask = x.gt(0)
        x = super().forward(x)
        return x.sum(dim=-2) / (
            (mask.sum(dim=-1) + tiny_value_of_dtype(torch.float)).unsqueeze(dim=-1)
        )



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/data/dataset_readers/dataset_reader.py
"""
import logging
from os import PathLike
from typing import Iterable, Iterator, Optional, Union, Dict, List

from overrides import overrides
from torch.utils.data import IterableDataset

from combo.config import FromParameters, Registry
from combo.config.from_parameters import register_arguments
from combo.data import LamboTokenizer, SingleIdTokenIndexer
from combo.data.instance import Instance
from combo.data.token_indexers import TokenIndexer
from combo.data.tokenizers import Tokenizer

logger = logging.getLogger(__name__)


PathOrStr = Union[PathLike, str]
DatasetReaderInput = Union[PathOrStr, List[PathOrStr], Dict[str, PathOrStr]]


@Registry.register('base_dataset_reader')
class DatasetReader(IterableDataset, FromParameters):
    """
    A `DatasetReader` knows how to turn a file containing a dataset into a collection
    of `Instance`s.
    """
    @register_arguments
    def __init__(self,
                 tokenizer: Optional[Tokenizer] = None,
                 token_indexers: Optional[Dict[str, TokenIndexer]] = None) -> None:
        super(DatasetReader).__init__()
        # self.__file_path = None
        self.__tokenizer = tokenizer or LamboTokenizer()
        self.__token_indexers = token_indexers or {"tokens": SingleIdTokenIndexer()}

    @property
    def tokenizer(self) -> Optional[Tokenizer]:
        return self.__tokenizer

    @tokenizer.setter
    def tokenizer(self, value):
        self.__tokenizer = value

    @property
    def token_indexers(self) -> Optional[Dict[str, TokenIndexer]]:
        return self.__token_indexers

    @token_indexers.setter
    def token_indexers(self, token_indexers: Dict[str, TokenIndexer]):
        self.__token_indexers = token_indexers

    @overrides
    def __getitem__(self, item, **kwargs) -> Instance:
        raise NotImplementedError

    @overrides
    def __iter__(self) -> Iterator[Instance]:
        """
        Returns an iterator of instances that can be read from the file path.
        """
        # TODO: add multiprocessing
        for instance in self._read():
            self.apply_token_indexers(instance)
            yield instance

    def _read(self, file_path: DatasetReaderInput) -> Iterable[Instance]:
        raise NotImplementedError

    def read(self, file_path: DatasetReaderInput) -> Iterator[Instance]:
        """
        Returns an iterator of instances that can be read from the file path.
        """
        for instance in self._read(file_path):  # type: ignore
            self.apply_token_indexers(instance)
            yield instance

    def apply_token_indexers(self, instance: Instance) -> None:
        pass

    def text_to_instance(self, *inputs) -> Instance:
        """
        Does whatever tokenization or processing is necessary to go from textual input to an
        `Instance`.  The primary intended use for this is with a
        :class:`~allennlp.predictors.predictor.Predictor`, which gets text input as a JSON
        object and needs to process it to be input to a model.

        The intent here is to share code between :func:`_read` and what happens at
        model serving time, or any other time you want to make a prediction from new data.  We need
        to process the data in the same way it was done at training time.  Allowing the
        `DatasetReader` to process new text lets us accomplish this, as we can just call
        `DatasetReader.text_to_instance` when serving predictions.

        The input type here is rather vaguely specified, unfortunately.  The `Predictor` will
        have to make some assumptions about the kind of `DatasetReader` that it's using, in order
        to pass it the right information.
        """
        raise NotImplementedError



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/predictors/predictor.py
"""

import json
import logging
import re
from contextlib import contextmanager
from typing import List, Iterator, Dict, Tuple, Any

import numpy
import pytorch_lightning as pl
import torch
from overrides import overrides
from torch import Tensor
from torch import backends
from torch.utils.hooks import RemovableHandle

from combo.common.util import sanitize
from combo.config import FromParameters
from combo.data.batch import Batch
from combo.data.dataset_readers.dataset_reader import DatasetReader
from combo.data.instance import JsonDict, Instance
from combo.modules.model import Model
from combo.nn import utils
from combo.nn.utils import move_to_device

logger = logging.getLogger(__name__)


class PredictorModule(pl.LightningModule, FromParameters):
    """
    a `Predictor` is a thin wrapper around an AllenNLP model that handles JSON -> JSON predictions
    that can be used for serving models through the web API or making predictions in bulk.
    """

    def __init__(self, model: Model, dataset_reader: DatasetReader, frozen: bool = True) -> None:
        super().__init__()
        if frozen:
            model.eval()
        self._model = model
        self.dataset_reader = dataset_reader
        self.cuda_device = next(self._model.named_parameters())[1].get_device()
        self._token_offsets: List[Tensor] = []

    def forward(self, inputs: Instance) -> Dict[str, torch.Tensor]:
        r"""
        Same as :meth:`torch.nn.Module.forward`.

        Args:
            *args: Whatever you decide to pass into the forward method.
            **kwargs: Keyword arguments are also possible.

        Return:
            Your model's output
        """
        return self._model.forward(inputs)

    @overrides(check_signature=False)
    def training_step(self, *args: Any, **kwargs: Any) -> Dict[str, torch.Tensor]:
        raise NotImplementedError()

    @overrides(check_signature=False)
    def configure_optimizers(self) -> Any:
        raise NotImplementedError()


    def load_line(self, line: str) -> JsonDict:
        """
        If your inputs are not in JSON-lines format (e.g. you have a CSV)
        you can override this function to parse them correctly.
        """
        return json.loads(line)

    def dump_line(self, outputs: Any) -> Any:
        """
        If you don't want your outputs in JSON-lines format
        you can override this function to output them differently.
        """
        return json.dumps(outputs) + "\n"

    def predict_json(self, inputs: JsonDict) -> Any:
        instance = self._json_to_instance(inputs)
        return self.predict_instance(instance)

    def json_to_labeled_instances(self, inputs: JsonDict) -> List[Instance]:
        """
        Converts incoming json to a [`Instance`](../data/instance.md),
        runs the model on the newly created instance, and adds labels to the
        `Instance`s given by the model's output.

        # Returns

        `List[instance]`
            A list of `Instance`'s.
        """

        instance = self._json_to_instance(inputs)
        outputs = self._model.forward_on_instance(instance)
        new_instances = self.predictions_to_labeled_instances(instance, outputs)
        return new_instances

    def get_gradients(self, instances: List[Instance]) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Gets the gradients of the loss with respect to the model inputs.

        # Parameters

        instances : `List[Instance]`

        # Returns

        `Tuple[Dict[str, Any], Dict[str, Any]]`
            The first item is a Dict of gradient entries for each input.
            The keys have the form  `{grad_input_1: ..., grad_input_2: ... }`
            up to the number of inputs given. The second item is the model's output.

        # Notes

        Takes a `JsonDict` representing the inputs of the model and converts
        them to [`Instances`](../data/instance.md)), sends these through
        the model [`forward`](../models/model.md#forward) function after registering hooks on the embedding
        layer of the model. Calls `backward` on the loss and then removes the
        hooks.
        """
        # set requires_grad to true for all parameters, but save original values to
        # restore them later
        original_param_name_to_requires_grad_dict = {}
        for param_name, param in self._model.named_parameters():
            original_param_name_to_requires_grad_dict[param_name] = param.requires_grad
            param.requires_grad = True

        embedding_gradients: List[Tensor] = []
        hooks: List[RemovableHandle] = self._register_embedding_gradient_hooks(embedding_gradients)

        dataset = Batch(instances)
        dataset.index_instances(self._model.vocab)
        dataset_tensor_dict = move_to_device(dataset.as_tensor_dict(), self.cuda_device)
        # To bypass "RuntimeError: cudnn RNN backward can only be called in training mode"
        with backends.cudnn.flags(enabled=False):
            outputs = self._model.make_output_human_readable(
                self._model.forward(**dataset_tensor_dict)  # type: ignore
            )

            loss = outputs["loss"]
            # Zero gradients.
            # NOTE: this is actually more efficient than calling `self._model.zero_grad()`
            # because it avoids a read op when the gradients are first updated below.
            for p in self._model.parameters():
                p.grad = None
            loss.backward()

        for hook in hooks:
            hook.remove()

        grad_dict = dict()
        for idx, grad in enumerate(embedding_gradients):
            key = "grad_input_" + str(idx + 1)
            grad_dict[key] = grad.detach().cpu().numpy()

        # restore the original requires_grad values of the parameters
        for param_name, param in self._model.named_parameters():
            param.requires_grad = original_param_name_to_requires_grad_dict[param_name]

        return grad_dict, outputs

    def get_interpretable_layer(self) -> torch.nn.Module:
        """
        Returns the input/embedding layer of the model.
        If the predictor wraps around a non-AllenNLP model,
        this function should be overridden to specify the correct input/embedding layer.
        For the cases where the input layer _is_ an embedding layer, this should be the
        layer 0 of the embedder.
        """
        try:
            return utils.find_embedding_layer(self._model)
        except RuntimeError:
            raise RuntimeError(
                "If the model does not use `TextFieldEmbedder`, please override "
                "`get_interpretable_layer` in your predictor to specify the embedding layer."
            )

    def get_interpretable_text_field_embedder(self) -> torch.nn.Module:
        """
        Returns the first `TextFieldEmbedder` of the model.
        If the predictor wraps around a non-AllenNLP model,
        this function should be overridden to specify the correct embedder.
        """
        try:
            return utils.find_text_field_embedder(self._model)
        except RuntimeError:
            raise RuntimeError(
                "If the model does not use `TextFieldEmbedder`, please override "
                "`get_interpretable_text_field_embedder` in your predictor to specify "
                "the embedding layer."
            )

    def _register_embedding_gradient_hooks(self, embedding_gradients):
        """
        Registers a backward hook on the embedding layer of the model.  Used to save the gradients
        of the embeddings for use in get_gradients()

        When there are multiple inputs (e.g., a passage and question), the hook
        will be called multiple times. We append all the embeddings gradients
        to a list.

        We additionally add a hook on the _forward_ pass of the model's `TextFieldEmbedder` to save
        token offsets, if there are any.  Having token offsets means that you're using a mismatched
        token indexer, so we need to aggregate the gradients across wordpieces in a token.  We do
        that with a simple sum.
        """

        def hook_layers(module, grad_in, grad_out):
            grads = grad_out[0]
            if self._token_offsets:
                # If you have a mismatched indexer with multiple TextFields, it's quite possible
                # that the order we deal with the gradients is wrong.  We'll just take items from
                # the list one at a time, and try to aggregate the gradients.  If we got the order
                # wrong, we should crash, so you'll know about it.  If you get an error because of
                # that, open an issue on github, and we'll see what we can do.  The intersection of
                # multiple TextFields and mismatched indexers is pretty small (currently empty, that
                # I know of), so we'll ignore this corner case until it's needed.
                offsets = self._token_offsets.pop(0)
                span_grads, span_mask = utils.batched_span_select(grads.contiguous(), offsets)
                span_mask = span_mask.unsqueeze(-1)
                span_grads *= span_mask  # zero out paddings

                span_grads_sum = span_grads.sum(2)
                span_grads_len = span_mask.sum(2)
                # Shape: (batch_size, num_orig_tokens, embedding_size)
                grads = span_grads_sum / torch.clamp_min(span_grads_len, 1)

                # All the places where the span length is zero, write in zeros.
                grads[(span_grads_len == 0).expand(grads.shape)] = 0

            embedding_gradients.append(grads)

        def get_token_offsets(module, inputs, outputs):
            offsets = utils.get_token_offsets_from_text_field_inputs(inputs)
            if offsets is not None:
                self._token_offsets.append(offsets)

        hooks = []
        text_field_embedder = self.get_interpretable_text_field_embedder()
        hooks.append(text_field_embedder.register_forward_hook(get_token_offsets))
        embedding_layer = self.get_interpretable_layer()
        hooks.append(embedding_layer.register_backward_hook(hook_layers))
        return hooks

    @contextmanager
    def capture_model_internals(self, module_regex: str = ".*") -> Iterator[dict]:
        """
        Context manager that captures the internal-module outputs of
        this predictor's model. The idea is that you could use it as follows:

        ```
            with predictor.capture_model_internals() as internals:
                outputs = predictor.predict_json(inputs)

            return {**outputs, "model_internals": internals}
        ```
        """
        results = {}
        hooks = []

        # First we'll register hooks to add the outputs of each module to the results dict.
        def add_output(idx: int):
            def _add_output(mod, _, outputs):
                results[idx] = {"name": str(mod), "output": sanitize(outputs)}

            return _add_output

        regex = re.compile(module_regex)
        for idx, (name, module) in enumerate(self._model.named_modules()):
            if regex.fullmatch(name) and module != self._model:
                hook = module.register_forward_hook(add_output(idx))
                hooks.append(hook)

        # If you capture the return value of the context manager, you get the results dict.
        yield results

        # And then when you exit the context we remove all the hooks.
        for hook in hooks:
            hook.remove()

    def predict_instance(self, instance: Instance) -> Any:
        outputs = self._model.forward_on_instance(instance)
        return sanitize(outputs)

    def predictions_to_labeled_instances(
        self, instance: Instance, outputs: Dict[str, numpy.ndarray]
    ) -> List[Instance]:
        """
        This function takes a model's outputs for an Instance, and it labels that instance according
        to the output. For example, in classification this function labels the instance according
        to the class with the highest probability. This function is used to to compute gradients
        of what the model predicted. The return type is a list because in some tasks there are
        multiple predictions in the output (e.g., in NER a model predicts multiple spans). In this
        case, each instance in the returned list of Instances contains an individual
        entity prediction as the label.
        """

        raise RuntimeError("implement this method for model interpretations or attacks")

    def _json_to_instance(self, json_dict: JsonDict) -> Instance:
        """
        Converts a JSON object into an [`Instance`](../data/instance.md)
        and a `JsonDict` of information which the `Predictor` should pass through,
        such as tokenised inputs.
        """
        raise NotImplementedError

    def predict_batch_json(self, inputs: List[JsonDict]) -> List[Any]:
        instances = self._batch_json_to_instances(inputs)
        return self.predict_batch_instance(instances)

    def predict_batch_instance(self, instances: List[Instance]) -> List[Any]:
        outputs = self._model.forward_on_instances(instances)
        return sanitize(outputs)

    def _batch_json_to_instances(self, json_dicts: List[JsonDict]) -> List[Instance]:
        """
        Converts a list of JSON objects into a list of `Instance`s.
        By default, this expects that a "batch" consists of a list of JSON blobs which would
        individually be predicted by `predict_json`. In order to use this method for
        batch prediction, `_json_to_instance` should be implemented by the subclass, or
        if the instances have some dependency on each other, this method should be overridden
        directly.
        """
        instances = []
        for json_dict in json_dicts:
            instances.append(self._json_to_instance(json_dict))
        return instances



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/main/allennlp/common/cached_transformers.py
"""
import logging
import re
import warnings

import transformers
from typing import Dict, Optional, Union, Tuple, NamedTuple

from transformers import AutoModel, AutoConfig

from combo.common.util import hash_object
from combo.utils import ConfigurationError, cast

logger = logging.getLogger(__name__)


class TransformerSpec(NamedTuple):
    model_name: str
    override_weights_file: Optional[str] = None
    override_weights_strip_prefix: Optional[str] = None
    reinit_modules: Optional[Union[int, Tuple[int, ...], Tuple[str, ...]]] = None


_tokenizer_cache: Dict[Tuple[str, str], transformers.PreTrainedTokenizer] = {}
_model_cache: Dict[TransformerSpec, transformers.PreTrainedModel] = {}

def get(
    model_name: str,
    make_copy: bool,
    override_weights_file: Optional[str] = None,
    override_weights_strip_prefix: Optional[str] = None,
    reinit_modules: Optional[Union[int, Tuple[int, ...], Tuple[str, ...]]] = None,
    load_weights: bool = True,
    **kwargs,
) -> transformers.PreTrainedModel:
    """
    Returns a transformer model from the cache.

    # Parameters

    model_name : `str`
        The name of the transformer, for example `"bert-base-cased"`
    make_copy : `bool`
        If this is `True`, return a copy of the model instead of the cached model itself. If you want to modify the
        parameters of the model, set this to `True`. If you want only part of the model, set this to `False`, but
        make sure to `copy.deepcopy()` the bits you are keeping.
    override_weights_file : `str`, optional (default = `None`)
        If set, this specifies a file from which to load alternate weights that override the
        weights from huggingface. The file is expected to contain a PyTorch `state_dict`, created
        with `torch.save()`.
    override_weights_strip_prefix : `str`, optional (default = `None`)
        If set, strip the given prefix from the state dict when loading it.
    reinit_modules: `Optional[Union[int, Tuple[int, ...], Tuple[str, ...]]]`, optional (default = `None`)
        If this is an integer, the last `reinit_modules` layers of the transformer will be
        re-initialized. If this is a tuple of integers, the layers indexed by `reinit_modules` will
        be re-initialized. Note, because the module structure of the transformer `model_name` can
        differ, we cannot guarantee that providing an integer or tuple of integers will work. If
        this fails, you can instead provide a tuple of strings, which will be treated as regexes and
        any module with a name matching the regex will be re-initialized. Re-initializing the last
        few layers of a pretrained transformer can reduce the instability of fine-tuning on small
        datasets and may improve performance (https://arxiv.org/abs/2006.05987v3). Has no effect
        if `load_weights` is `False` or `override_weights_file` is not `None`.
    load_weights : `bool`, optional (default = `True`)
        If set to `False`, no weights will be loaded. This is helpful when you only
        want to initialize the architecture, like when you've already fine-tuned a model
        and are going to load the weights from a state dict elsewhere.
    """
    global _model_cache
    spec = TransformerSpec(
        model_name,
        override_weights_file,
        override_weights_strip_prefix,
        reinit_modules,
    )
    transformer = _model_cache.get(spec, None)
    if transformer is None:
        if not load_weights:
            if override_weights_file is not None:
                warnings.warn(
                    "You specified an 'override_weights_file' in allennlp.common.cached_transformers.get(), "
                    "but 'load_weights' is set to False, so 'override_weights_file' will be ignored.",
                    UserWarning,
                )
            if reinit_modules is not None:
                warnings.warn(
                    "You specified 'reinit_modules' in allennlp.common.cached_transformers.get(), "
                    "but 'load_weights' is set to False, so 'reinit_modules' will be ignored.",
                    UserWarning,
                )
            transformer = AutoModel.from_config(
                AutoConfig.from_pretrained(
                    model_name,
                    **kwargs,
                )
            )
        elif override_weights_file is not None:
            if reinit_modules is not None:
                warnings.warn(
                    "You specified 'reinit_modules' in allennlp.common.cached_transformers.get(), "
                    "but 'override_weights_file' is not None, so 'reinit_modules' will be ignored.",
                    UserWarning,
                )
            import torch
            from combo.utils.file_utils import cached_path

            override_weights_file = cached_path(override_weights_file)
            override_weights = torch.load(override_weights_file)
            if override_weights_strip_prefix is not None:

                def strip_prefix(s):
                    if s.startswith(override_weights_strip_prefix):
                        return s[len(override_weights_strip_prefix) :]
                    else:
                        return s

                valid_keys = {
                    k
                    for k in override_weights.keys()
                    if k.startswith(override_weights_strip_prefix)
                }
                if len(valid_keys) > 0:
                    logger.info(
                        "Loading %d tensors from %s", len(valid_keys), override_weights_file
                    )
                else:
                    raise ValueError(
                        f"Specified prefix of '{override_weights_strip_prefix}' means no tensors "
                        f"will be loaded from {override_weights_file}."
                    )
                override_weights = {strip_prefix(k): override_weights[k] for k in valid_keys}

            transformer = AutoModel.from_config(
                AutoConfig.from_pretrained(
                    model_name,
                    **kwargs,
                )
            )
            # When DistributedDataParallel or DataParallel is used, the state dict of the
            # DistributedDataParallel/DataParallel wrapper prepends "module." to all parameters
            # of the actual model, since the actual model is stored within the module field.
            # This accounts for if a pretained model was saved without removing the
            # DistributedDataParallel/DataParallel wrapper.
            if hasattr(transformer, "module"):
                transformer.module.load_state_dict(override_weights)
            else:
                transformer.load_state_dict(override_weights)
        elif reinit_modules is not None:
            transformer = AutoModel.from_pretrained(
                model_name,
                **kwargs,
            )
            num_layers = transformer.config.num_hidden_layers
            if isinstance(reinit_modules, int):
                reinit_modules = tuple(range(num_layers - reinit_modules, num_layers))
            if all(isinstance(x, int) for x in reinit_modules):
                # This type cast is neccessary to avoid a mypy error.
                reinit_modules = cast(Tuple[int], reinit_modules)
                if any(layer_idx < 0 or layer_idx > num_layers for layer_idx in reinit_modules):
                    raise ValueError(
                        f"A layer index in reinit_modules ({reinit_modules}) is invalid."
                        f" Must be between 0 and the maximum layer index ({num_layers - 1}.)"
                    )
                # Some transformer models organize their modules differently, so if this fails,
                # raise an error with a helpful message.
                try:
                    for layer_idx in reinit_modules:
                        transformer.encoder.layer[layer_idx].apply(transformer._init_weights)
                except AttributeError:
                    raise ConfigurationError(
                        f"Unable to re-initialize the layers of transformer model"
                        f" {model_name} using layer indices. Please provide a tuple of"
                        " strings corresponding to the names of the layers to re-initialize."
                    )
            elif all(isinstance(x, str) for x in reinit_modules):
                for regex in reinit_modules:
                    for name, module in transformer.named_modules():
                        if re.search(str(regex), name):
                            module.apply(transformer._init_weights)
            else:
                raise ValueError(
                    "reinit_modules must be either an integer, a tuple of strings, or a tuple of integers."
                )
        else:
            transformer = AutoModel.from_pretrained(
                model_name,
                **kwargs,
            )
        _model_cache[spec] = transformer
    if make_copy:
        import copy

        return copy.deepcopy(transformer)
    else:
        return transformer

def get_tokenizer(model_name: str, **kwargs) -> transformers.PreTrainedTokenizer:

    cache_key = (model_name, hash_object(kwargs))

    global _tokenizer_cache
    tokenizer = _tokenizer_cache.get(cache_key, None)
    if tokenizer is None:
        tokenizer = transformers.AutoTokenizer.from_pretrained(
            model_name,
            **kwargs,
        )
        _tokenizer_cache[cache_key] = tokenizer
    return tokenizer



"""
Adapted from AllenNLP
https://github.com/allenai/allennlp/blob/80fb6061e568cb9d6ab5d45b661e86eb61b92c82/allennlp/common/file_utils.py#L82
"""
import os
from os import PathLike
from pathlib import Path
from typing import Union
import cached_path as _cached_path


CACHE_ROOT = Path(os.getenv("COMBO_CACHE_ROOT", Path.home() / ".combo"))
CACHE_DIRECTORY = str(CACHE_ROOT / "cache")


def get_file_extension(path: str, dot=True, lower: bool = True):
    ext = os.path.splitext(path)[1]
    ext = ext if dot else ext[1:]
    return ext.lower() if lower else ext


def cached_path(
    url_or_filename: Union[str, PathLike],
    cache_dir: Union[str, Path] = None,
    extract_archive: bool = False,
    force_extract: bool = False,
) -> str:
    """
    Given something that might be a URL or local path, determine which.
    If it's a remote resource, download the file and cache it, and
    then return the path to the cached file. If it's already a local path,
    make sure the file exists and return the path.
    For URLs, "http://", "https://", "s3://", "gs://", and "hf://" are all supported.
    The latter corresponds to the HuggingFace Hub.
    For example, to download the PyTorch weights for the model `epwalsh/bert-xsmall-dummy`
    on HuggingFace, you could do:
    ```python
    cached_path("hf://epwalsh/bert-xsmall-dummy/pytorch_model.bin")
    ```
    For paths or URLs that point to a tarfile or zipfile, you can also add a path
    to a specific file to the `url_or_filename` preceeded by a "!", and the archive will
    be automatically extracted (provided you set `extract_archive` to `True`),
    returning the local path to the specific file. For example:
    ```python
    cached_path("model.tar.gz!weights.th", extract_archive=True)
    ```
    # Parameters
    url_or_filename : `Union[str, Path]`
        A URL or path to parse and possibly download.
    cache_dir : `Union[str, Path]`, optional (default = `None`)
        The directory to cache downloads.
    extract_archive : `bool`, optional (default = `False`)
        If `True`, then zip or tar.gz archives will be automatically extracted.
        In which case the directory is returned.
    force_extract : `bool`, optional (default = `False`)
        If `True` and the file is an archive file, it will be extracted regardless
        of whether or not the extracted directory already exists.
        !!! Warning
            Use this flag with caution! This can lead to race conditions if used
            from multiple processes on the same file.
    """
    return str(
        _cached_path.cached_path(
            url_or_filename,
            cache_dir=cache_dir or CACHE_DIRECTORY,
            extract_archive=extract_archive,
            force_extract=force_extract,
        )
    )


"""
Adapted from COMBO
Author: Mateusz Klimaszewski
"""

import torch
import os


class ConfigurationError(Exception):
    def __init__(self, message: str):
        super().__init__()
        self.message = message


def file_exists(*paths):
    """Check whether paths exists."""
    for path in paths:
        if path is None:
            raise ConfigurationError("File cannot be None")
        if not os.path.exists(path):
            raise ConfigurationError(f"Could not find the file at path: '{path}'")


def check_size_match(size_1: torch.Size, size_2: torch.Size, tensor_1_name: str, tensor_2_name: str):
    """Check if tensors' sizes are the same."""
    if size_1 != size_2:
        raise ConfigurationError(
            f"{tensor_1_name} must match {tensor_2_name}, but got {size_1} "
            f"and {size_2} instead"
        )


import collections
import dataclasses
import json
from dataclasses import dataclass, field
from typing import List, Dict, Any

import conllu
from conllu.models import Metadata
from overrides import overrides

from combo.data.tokenizers import Token


@dataclass
class Sentence:
    tokens: List[Token] = field(default_factory=list)
    sentence_embedding: List[float] = field(default_factory=list, repr=False)
    relation_distribution: List[float] = field(default_factory=list, repr=False)
    relation_label_distribution: List[float] = field(default_factory=list, repr=False)
    metadata: Dict[str, Any] = field(default_factory=collections.OrderedDict)

    def to_json(self):
        return json.dumps({
            "tokens": [dataclasses.asdict(t) for t in self.tokens],
            "sentence_embedding": self.sentence_embedding,
            "relation_distribution": self.relation_distribution,
            "relation_label_distribution": self.relation_label_distribution,
            "metadata": self.metadata
        })

    def __len__(self):
        return len(self.tokens)


class _TokenList(conllu.models.TokenList):

    @overrides
    def __repr__(self):
        return 'TokenList<' + ', '.join(token['text'] for token in self) + '>'


def sentence2conllu(sentence: Sentence, keep_semrel: bool = True) -> conllu.models.TokenList:
    tokens = []
    for token in sentence.tokens:
        token_dict = collections.OrderedDict(token.as_dict(keep_semrel))
        tokens.append(token_dict)
    # Range tokens must be tuple not list, this is conllu library requirement
    for t in tokens:
        if type(t["idx"]) == list:
            t["idx"] = tuple(t["idx"])
        if t["deps"]:
            for dep in t["deps"]:
                if len(dep) > 1 and type(dep[1]) == list:
                    dep[1] = tuple(dep[1])
    return _TokenList(tokens=tokens,
                      metadata=sentence.metadata if sentence.metadata is None else Metadata())

def serialize_field(field: Any) -> str:
    if field is None:
        return '_'

    if isinstance(field, dict):
        if field == {}:
            return '_'

        fields = []
        for key, value in field.items():
            if value is None:
                value = "_"
            if value == "":
                fields.append(key)
                continue

            fields.append('='.join((key, value)))

        return '|'.join(fields)

    if isinstance(field, tuple):
        return "".join([serialize_field(item) for item in field])

    if isinstance(field, list):
        if len(field[0]) != 2:
            raise ValueError("Can't serialize '{}', invalid format".format(field))
        return "|".join([serialize_field(value) + ":" + str(key) for key, value in field])

    return "{}".format(field)

def serialize_token_list(tokenlist: conllu.models.TokenList) -> str:
    KEYS_ORDER = ['idx', 'text', 'lemma', 'upostag', 'xpostag', 'feats', 'head', 'deprel', 'deps']
    lines = []

    if tokenlist.metadata:
        for key, value in tokenlist.metadata.items():
            if value:
                line = f"# {key} = {value}"
            else:
                line = f"# {key}"
            lines.append(line)

    for token_data in tokenlist:
        line = '\t'.join(serialize_field(token_data[k]) for k in KEYS_ORDER)
        serialized_misc = serialize_field(token_data['misc'])
        serialized_entity_type = serialize_field(token_data['entity_type'])
        if serialized_misc == '_' and serialized_entity_type == '_':
            serialized_last_column = '_'
        elif serialized_misc == '_':
            serialized_last_column = serialized_entity_type
        elif serialized_entity_type == '_':
            serialized_last_column = serialized_misc
        else:
            serialized_last_column = serialized_entity_type + ' | ' + serialized_misc

        line += '\t' + serialized_last_column
        lines.append(line)

    return '\n'.join(lines) + "\n\n"

def tokens2conllu(tokens: List[str]) -> conllu.models.TokenList:
    return _TokenList(
        [collections.OrderedDict({"id": idx, "token": token}) for
         idx, token
         in enumerate(tokens, start=1)],
        metadata=collections.OrderedDict()
    )


def conllu2sentence(tokens: List[Token],
                    sentence_embedding=None, embeddings=None,
                    metadata=None,
                    relation_distribution=None,
                    relation_label_distribution=None) -> Sentence:
    embeddings = embeddings or {}
    if relation_distribution is None:
        relation_distribution = []
    if relation_label_distribution is None:
        relation_label_distribution = []
    if sentence_embedding is None:
        sentence_embedding = []
    if embeddings:
        for token in tokens:
            token.embeddings = embeddings[token["idx"]]
    return Sentence(
        tokens=tokens,
        sentence_embedding=sentence_embedding,
        relation_distribution=relation_distribution,
        relation_label_distribution=relation_label_distribution,
        metadata=metadata
    )
